# Copyright 2013 tsuru-circus authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import unittest

from tsuru.stream import extract_message


class ExtractMessageTestCase(unittest.TestCase):
    def test_extract_message_when_it_isnt_a_list(self):
        msg = 'Starting gunicorn 0.15.0'
        expected = 'Starting gunicorn 0.15.0'
        result = extract_message(msg)
        self.assertListEqual([expected], result)

    def test_extract_message(self):
        msg = '2012-11-06 17:13:55 [12019] [INFO] Starting gunicorn 0.15.0'
        expected = 'Starting gunicorn 0.15.0'
        result = extract_message(msg)
        self.assertListEqual([expected], result)

    def test_extract_multiline(self):
        msg = "2012-11-06 18:30:10 [13887] [INFO] Listening at: http://127.0.0.1:8000 (13887)\n2012-11-06 18:30:10 [13887] [INFO] Using worker: sync\n2012-11-06 18:30:10 [13890] [INFO] Booting worker with pid: 13890\n2012-11-06 18:30:10 [13890] [ERROR] Exception in worker process:\nTraceback (most recent call last):\n"
        expected = [
            "Listening at: http://127.0.0.1:8000 (13887)\n",
            "Using worker: sync\n",
            "Booting worker with pid: 13890\n",
            "Exception in worker process:\nTraceback (most recent call last):\n",
        ]
        result = extract_message(msg)
        self.assertListEqual(expected, result)



# Copyright 2013 tsuru-circus authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import copy
import os
import unittest

from tsuru.plugins import ApprcWatcher

NOPATH_OUTPUT = {
    "status": "ok",
    "options": {
        "max_retry": 5,
        "env": {"SOMETHING": "adios"},
    },
}


def remove_env(name):
    del os.environ[name]


def create_fake_call(status_return, options_return=NOPATH_OUTPUT):
    kw = []

    def fake_call(command, *args, **kwargs):
        if command == "status":
            return status_return
        elif command == "set":
            kw.append(kwargs)
            return {"set": "ok"}
        elif command == "options":
            return options_return
    return fake_call, kw


class ApprcWatcherTest(unittest.TestCase):
    def test_add_envs(self):
        plugin = ApprcWatcher("", "", 1)
        plugin.call, kw = create_fake_call(None)
        plugin.add_envs(name="name", envs={"foo": "bar"})
        env = copy.deepcopy(os.environ)
        env["foo"] = "bar"
        expected = [{"name": "name", "options": {"env": env}}]
        self.assertEqual(expected, kw)

    def test_add_envs_may_override_os_environ(self):
        os.environ["SOMETHING_UNKNOWN"] = "123"
        self.addCleanup(remove_env, "SOMETHING_UNKNOWN")
        plugin = ApprcWatcher("", "", 1)
        plugin.call, kw = create_fake_call(None)
        plugin.add_envs(name="name", envs={"foo": "bar", "SOMETHING_UNKNOWN": "456"})
        env = copy.deepcopy(os.environ)
        env["SOMETHING_UNKNOWN"] = "456"
        env["foo"] = "bar"
        expected = [{"name": "name", "options": {"env": env}}]
        self.assertEqual(expected, kw)

    def test_add_envs_dont_call_set_when_variables_dont_change(self):
        plugin = ApprcWatcher("", "", 1)
        envs = copy.deepcopy(NOPATH_OUTPUT)
        envs["options"]["env"].update(os.environ)
        plugin.call, kw = create_fake_call(None, envs)
        plugin.add_envs(name="name", envs=NOPATH_OUTPUT["options"]["env"])
        self.assertEqual([], kw)

    def test_add_envs_dont_call_set_when_variables_dont_change_and_ses_os_environ(self):
        plugin = ApprcWatcher("", "", 1)
        envs = copy.deepcopy(NOPATH_OUTPUT)
        envs["options"]["env"].update(os.environ)
        plugin.call, kw = create_fake_call(None, envs)
        plugin.add_envs(name="name", envs=envs["options"]["env"])
        self.assertEqual([], kw)

    def test_add_envs_override_environs(self):
        plugin = ApprcWatcher("", "", 1)
        envs = copy.deepcopy(NOPATH_OUTPUT)
        envs["options"]["env"]["foo"] = "foo"
        plugin.call, kw = create_fake_call(None, envs)
        plugin.add_envs(name="name", envs={"foo": "bar"})
        env = copy.deepcopy(os.environ)
        env["foo"] = "bar"
        expected = [{"name": "name", "options": {"env": env}}]
        self.assertEqual(expected, kw)

    def test_add_env_ignores_PYTHONPATH_from_os_environ(self):
        os.environ["PYTHONPATH"] = "/home/user/python"
        self.addCleanup(remove_env, "PYTHONPATH")
        plugin = ApprcWatcher("", "", 1)
        plugin.call, kw = create_fake_call(None)
        plugin.add_envs(name="name", envs={"foo": "bar"})
        env = copy.deepcopy(os.environ)
        env["foo"] = "bar"
        del env["PYTHONPATH"]
        expected = [{"name": "name", "options": {"env": env}}]
        self.assertEqual(expected, kw)

    def test_add_env_doesnt_ignore_PYTHONPATH_from_apprc(self):
        plugin = ApprcWatcher("", "", 1)
        plugin.call, kw = create_fake_call(None)
        plugin.add_envs(name="name", envs={"foo": "bar", "PYTHONPATH": "/usr/lib/python"})
        env = copy.deepcopy(os.environ)
        env["foo"] = "bar"
        env["PYTHONPATH"] = "/usr/lib/python"
        expected = [{"name": "name", "options": {"env": env}}]
        self.assertEqual(expected, kw)

    def test_reload_env_add_envs(self):
        sr = {"statuses": {"name": "name", "cmd": "cmd"}}
        fake_call, kw = create_fake_call(sr)
        plugin = ApprcWatcher("", "", 1)
        plugin.call = fake_call
        plugin.apprc = os.path.join(os.path.dirname(__file__),
                                    "testdata/apprc")
        plugin.reload_env()
        env = {"VAR1": "value-1", "port": "8888", "VAR2": "value2"}
        env.update(os.environ)
        expected = [
            {"name": "cmd", "options": {"env": env}},
            {"name": "name", "options": {"env": env}},
        ]
        self.assertEqual(expected, sorted(kw, key=lambda x: x["name"]))

    def test_reload_env_skips_plugins(self):
        sr = {"statuses": {"name": "name",
                           "plugin:tsuru-circus-ApprcWatcher": "ok"}}
        fake_call, kw = create_fake_call(sr)
        plugin = ApprcWatcher("", "", 1)
        plugin.call = fake_call
        plugin.apprc = os.path.join(os.path.dirname(__file__),
                                    "testdata/apprc")
        plugin.reload_env()
        env = {"VAR1": "value-1", "port": "8888", "VAR2": "value2"}
        env.update(os.environ)
        expected = [
            {"name": "name", "options": {"env": env}},
        ]
        self.assertEqual(expected, kw)



# Copyright 2013 tsuru-circus authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import json
import os

from circus.plugins import CircusPlugin
from circus.client import CircusClient
from honcho.procfile import Procfile
from zmq.eventloop import ioloop

from tsuru import common


class FileWatcher(object):
    def __init__(self, filename, callback):
        self._filename = filename
        self._callback = callback
        self._mtime = 0

    def __call__(self, *args, **kwargs):
        if os.path.exists(self._filename):
            mtime = os.path.getmtime(self._filename)
            if mtime > self._mtime:
                self._mtime = mtime
                self._callback()


class ApprcWatcher(CircusPlugin):
    name = "apprc_watcher"

    def __init__(self, *args, **config):
        super(ApprcWatcher, self).__init__(*args, **config)
        self.loop_rate = config.get("loop_rate", 3)  # in seconds
        self.apprc = config.get("apprc", "/home/application/apprc")
        self.port = config.get("port", "8888")
        self.period = ioloop.PeriodicCallback(FileWatcher(self.apprc, self.reload_env),
                                              self.loop_rate * 1000,
                                              self.loop)

    def handle_init(self):
        self.period.start()

    def handle_stop(self):
        self.period.stop()

    def handle_recv(self, data):
        pass

    def reload_env(self):
        envs = {"port": self.port}
        envs.update(common.load_envs(self.apprc))
        for name in self.cmds():
            if not name.startswith("plugin:"):
                self.add_envs(name, envs)

    def add_envs(self, name, envs):
        current = self.call("options", name=name)["options"]["env"]
        values = {}
        values.update(os.environ, **envs)
        if "PYTHONPATH" in values and "PYTHONPATH" not in envs:
            del values["PYTHONPATH"]
        if values != current:
            self.call("set", name=name, options={"env": values})

    def cmds(self):
        return self.call("status")["statuses"].keys()


class ProcfileWatcher(CircusPlugin):
    name = "procfile_watcher"

    def __init__(self, *args, **config):
        super(ProcfileWatcher, self).__init__(*args, **config)
        self.loop_rate = config.get("loop_rate", 3)  # in seconds
        self.procfile_path = config.get("app_path",
                                        "/home/application/current/Procfile")
        self.working_dir = config.get("working_dir",
                                      "/home/application/current")
        self.apprc = config.get("apprc", "/home/application/apprc")
        self.port = config.get("port", "8888")
        self.uid = config.get("uid", "ubuntu")
        self.stderr_stream = {"class": config.get("stderr_stream",
                                                  "tsuru.stream.Stream")}
        self.stdout_stream = {"class": config.get("stdout_stream",
                                                  "tsuru.stream.Stream")}
        file_watcher = FileWatcher(self.procfile_path, self.reload_procfile)
        self.period = ioloop.PeriodicCallback(file_watcher,
                                              self.loop_rate * 1000,
                                              self.loop)
        self.circus_client = CircusClient()

    def get_cmd(self, name):
        return self.call("get", name=name, keys=["cmd"])["options"]["cmd"]

    def handle_init(self):
        self.period.start()

    def handle_stop(self):
        self.period.stop()

    def handle_recv(self, data):
        pass

    def add_watcher(self, name, cmd):
        env = {"port": self.port}
        env.update(common.load_envs(self.apprc))
        options = {
            "env": env,
            "copy_env": True,
            "working_dir": self.working_dir,
            "stderr_stream": self.stderr_stream,
            "stdout_stream": self.stdout_stream,
            "uid": self.uid,
        }
        self.circus_client.call(json.dumps({
            "command": "add",
            "properties": {
                "cmd": cmd,
                "name": name,
                "args": [],
                "options": options,
                "start": True,
            },
        }))

    def remove_watcher(self, name):
        self.call("rm", name=name)

    def change_cmd(self, name, cmd):
        self.call("set", name=name, options={"cmd": cmd})

    def commands(self, procfile):
        cmds = self.call("status")["statuses"]
        cmds_names = set([k for k in cmds.keys() if not k.startswith("plugin:")])
        new_cmds = set(procfile.commands.keys())
        to_remove = cmds_names.difference(new_cmds)
        to_add = new_cmds.difference(cmds_names)
        to_change_names = cmds_names.intersection(new_cmds)
        to_change = {}
        for name in to_change_names:
            if self.get_cmd(name) != procfile.commands.get(name):
                to_change[name] = procfile.commands[name]
        return to_add, to_remove, to_change

    def reload_procfile(self):
        with open(self.procfile_path) as file:
            procfile = Procfile(file.read())
            to_add, to_remove, to_change = self.commands(procfile)

            for name in to_remove:
                self.remove_watcher(name)

            for name in to_add:
                self.add_watcher(name=name, cmd=procfile.commands[name])

            for name, cmd in to_change.items():
                self.change_cmd(name=name, cmd=procfile.commands[name])



# Copyright 2013 tsuru-circus authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import json
import re
import requests

from tsuru import common


def extract_message(msg):
    # 2012-11-06 18:30:10 [13887] [INFO]
    regex = "\d+\-\d+\-\d+ \d+\:\d+\:\d+ \[\d+\] \[\w+\] "
    msgs = re.split(regex, msg)
    return [m for m in msgs if m]


class Stream(object):

    def __init__(self, **kwargs):
        self.apprc = "/home/application/apprc"

    def __call__(self, data):
        appname, host, token = self.load_envs()
        if appname and host and token:
            url = "{0}/apps/{1}/log".format(host, appname)
            messages = extract_message(data["data"])
            requests.post(url, data=json.dumps(messages),
                          headers={"Authorization": "bearer " + token})

    def load_envs(self):
        envs = common.load_envs(self.apprc)
        return envs.get("TSURU_APPNAME"), envs.get("TSURU_HOST"), envs.get("TSURU_APP_TOKEN")

    def close(self):
        pass



# Copyright 2013 tsuru-circus authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import unittest
import mock
import os
import json

from tsuru.stream import Stream


class StreamTestCase(unittest.TestCase):
    def setUp(self):
        l = '2012-11-06 17:13:55 [12019] [INFO] Starting gunicorn 0.15.0\n'
        self.data = {
            'pid': 12018,
            'data': l,
            'name': 'stderr'
        }
        self.stream = Stream()
        self.stream.apprc = os.path.join(os.path.dirname(__file__),
                                         "testdata/apprc")

    def test_should_have_the_close_method(self):
        self.assertTrue(hasattr(Stream, "close"))

    @mock.patch("requests.post")
    def test_should_send_log_to_tsuru(self, post):
        post.return_value = mock.Mock(status_code=200)
        self.stream(self.data)
        appname, host, token = self.stream.load_envs()
        url = "{0}/apps/{1}/log".format(host, appname)
        expected_msg = "Starting gunicorn 0.15.0\n"
        expected_data = json.dumps([expected_msg])
        post.assert_called_with(url, data=expected_data,
                                headers={"Authorization": "bearer " + token})

    @mock.patch("tsuru.common.load_envs")
    def test_should_slience_errors_when_envs_does_not_exist(lenvs, self):
        lenvs.return_value = {}
        try:
            stream = Stream()
            stream(self.data)
        except Exception as e:
            self.fail("Should not fail when envs does not exist. Exception: %s" % e)



# Copyright 2013 tsuru-circus authors. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import os
import unittest

from tsuru.plugins import FileWatcher


class FileWatcherTest(unittest.TestCase):

    def test_invokes_callback(self):
        calls = {"callback": 0}

        def callback():
            calls["callback"] += 1

        path = os.path.join(os.path.dirname(__file__),
                            "testdata/apprc")
        watcher = FileWatcher(path, callback)
        watcher()
        self.assertEqual(1, calls["callback"])
        watcher()
        self.assertEqual(1, calls["callback"])
        os.system("touch %s" % path)
        watcher()
        self.assertEqual(2, calls["callback"])



import os
import unittest

from tsuru import common


class CommonTestCase(unittest.TestCase):

    def test_load_envs(self):
        path = os.path.join(os.path.dirname(__file__), "testdata", "apprc")
        want = {
            "VAR1": "value-1",
            "VAR2": "value2",
            "TSURU_APPNAME": "appname1",
            "TSURU_HOST": "host1",
        }
        got = common.load_envs(path)
        self.assertEqual(want, got)

    def test_load_envs_unknown_file(self):
        path = "/something/that/user/does/not/know"
        self.assertEqual({}, common.load_envs(path))
